#ifndef MATFILE_H
#define MATFILE_H
/* 20 bytes header:
(1)long type:
The type flag contains an integer whose decimal digits encode storage information. If the integer is represented as MOPT where M is the thousands digit, O is the hun dreds digit, P is the tens digit, and T is the ones digit, then:

M indicates the numeric format of binary numbers on the machine that wrote the file. Use this table to determine the number to use for your machine:
O	IEEE Little Endian (PC, 386, 486, DEC Rise)
1	IEEE Big Endian (Macintosh, SPARC, Apollo, SGI, HP 9000/300, other Motorola)
2	VAX D-float
3	VAX G-float
4	Cray

O is always O (zero) and is reserved for future use.

P indicates which format the data is stored in according to the following table:
0	double-precision (64-Bit) floating point numbers
1	single-precision (32-bit) floating point numbers
2	32-bit signed integers
3	16-bit signed integers
4	16-bit un signed integers
5 	8-bit unsigned integers

The precision used by the save command depends on the size and type of each Matrix. Matrices with any noninteger entries and Matrices with 10,000 or fewer elements are saved in floating point formats requiring eight bytes per real element. Matrices with all integer entries and more than 10,000 elements are saved in the following formats, requiring fewer bytes per element.
Element range	Bytes per element
	[0:255]			1
	f0:65535]		2
	[-32767 :32767]		2
	[-2^31+1:2^31-11	4
	other			8

T indicates the Matrix type according to the following tabled
O	Numeric (Full) Matrix
1	Text Matrix
2	Sparse Matrix

Note that the elements of a text Matrix are stored as floating point numbers between O and 255 representing ASCII-encoded characters.

(2)long mrows	
The row dimension contains an integer with the number of rows in the Matrix.

(3)long ncols
The column dimension contains an integer with thenumber of columns in the Matrix.

(4)long imag	
The imaginary flag is an integer whose value is either 0 or 1. If 1, then the Matrix has an imaginary part. If 0, there is only real data.

(5)long namlen	
The name length contains an integer with 1 plus the length of the Matrix name.
*/

/* this API only supports all the elements are stored in double format, do
   not distinguish for integer matrix */

#define MACHINE_ENDIAN 0
typedef struct {
        long    type;   /*      type */
        long    nrows;  /*      row dimension */
        long    ncols;  /*      colUmn dimension */
        long    imagf;  /*      flag indicating imag part*/
        long    namlen; /*      name length (including NULL) */
} Fmatrix;


extern PyArrayObject *transpose(PyArrayObject *ap);
extern int writeMatFile(FILE *fp, PyArrayObject *ap, const char *name);
extern PyArrayObject *readMatFile(FILE *fp, const char *varname);
extern void rbo(char * data, int bpe, int nel);

#endif
