#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<fcntl.h>
#include<string.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/wait.h>
#include<signal.h>
#include<time.h>
#include"Python.h"
#include"arrayobject.h"

#define OCT_ENGINE
#include"oct-engine.h"
#include"matfile.h"
static int echo = 0; /* whether to print out the eval echos from octave */
 
/* keep read until a terminator "> " is found or newline */
static char *myfgets(char *s, int size, int fd)
{
    int i, r;
     
    for (i = 0; i < size-1; i++) {
	r = read(fd, &s[i], 1);
	if ( r == -1) return NULL; /* error read */
	if ( r == 0 ) {
	    s[i] = '\0';
	    return s;
	}
	if ( s[i]=='\n' ) break;
	if ( s[i]==' ' && i>0 && s[i-1]=='>' ) break;
    }
    s[i+1] = '\0';
    return s;
}

extern Engine *engOpen(const char *startcmd)
{
    int cmdfd[2], resfd[2];
    pid_t childpid; 
    char tmp[BUF_SIZE], fifo[BUF_SIZE], ch;
    Engine *ep;
    
    if (pipe(cmdfd)!=0) { 
	perror("pipe");
	return NULL;
    }
    if (pipe(resfd)!=0) { 
	perror("pipe");
	return NULL;
    }
    for (ch=48; ch<58; ch++) {
	fifo[0] = '\0';
	strcpy(fifo, FIFOBASE);
	strncat(fifo, &ch, 1);
	if (mkfifo(fifo, 0666) == 0) break;
	if (ch == 57) {
	    perror("mkfifo");
	    printf("Please clear up the abandoned fifos %s*\n",
		   FIFOBASE);
	    return NULL;
	}
    }

    ep = (Engine *)malloc(sizeof(Engine));
    strcpy(ep->fifo, fifo);
    
    if ((childpid=fork()) == 0) { /* octave is the child */
	if (dup2(cmdfd[0], STDIN_FILENO)==-1) perror("dup2");
	if (dup2(resfd[1], STDOUT_FILENO)==-1) perror("dup2");
	close(cmdfd[0]);
	close(cmdfd[1]);
	close(resfd[0]);
	close(resfd[1]);
	if (execl("/usr/bin/octave", "octave", "--traditional", "-q", "-i", NULL) == -1)
	    perror("The exec of octave failed");
    } else {
	close(cmdfd[0]);
	close(resfd[1]);
	ep->cmd = cmdfd[1];
	ep->res = resfd[0];
	ep->out_buf = NULL;
	ep->out_buf_size = 0;
	    
	printf("Wait for Octave to startup ...\n");
	while (1) {
	    if ( myfgets(tmp, BUF_SIZE, ep->res) == NULL ) {
		perror("myfgets");
		free(ep);
		return NULL;
	    }
	    /* printf("%s", tmp); */
	    if ( strncmp(tmp, PROMPT, strlen(PROMPT)) == 0 ) break;
	}
	printf("Octave ready.\n\n");
	return ep;
    }
}

extern int engClose(Engine *ep)
{
    if (engEvalString(ep, "exit") != 0) return -1;
    printf("Wait for Octave to stop ...\n");
    sleep(2);			/* wait for the Octave to exit */
    close(ep->cmd);
    close(ep->res);
    if (ep->out_buf != NULL) free(ep->out_buf);
    unlink(ep->fifo);
    free(ep);
    return 0;
}

extern int engEvalString(Engine	*ep, const char *string)
{
    int sz, rd=0, newrd, rtn;
    char tmp[BUF_SIZE];

/*      printf("Cmd:  %s\n", string); */
    sz = strlen(string);
    if (sz > BUF_SIZE-2) {
	perror("Command too long");
	return 1;
    }
    strcpy(tmp, string);
    strcat(tmp, "\n");
    rtn = write(ep->cmd, tmp, sz+1);
    rtn = rtn - sz - 1;		/* always <= 0 */
    if ( strcmp(string, "exit") == 0 ) return rtn;
    
    /* deal with output */
    if (ep->out_buf == NULL) {	
	while (1) {
	    if ( myfgets(tmp, BUF_SIZE, ep->res) == NULL ) {
		perror("myfgets");
		return 1;
	    }
	    if ( (int)strlen(tmp) >= (int)strlen(PROMPT) 
		 && strncmp(tmp, PROMPT, strlen(PROMPT)) == 0 ) break;
	}
    } else {
	ep->out_buf[0] = '\0';
	while (1) {
	    if ( myfgets(tmp, BUF_SIZE, ep->res) == NULL ) {
		perror("myfgets");
		return 1;
	    }
	    if ( (int)strlen(tmp) >= (int)strlen(PROMPT) 
		 && strncmp(tmp, PROMPT, strlen(PROMPT)) == 0 ) break;
	    if (rd < ep->out_buf_size-1) {
		newrd = MIN(ep->out_buf_size - rd, (int)strlen(tmp));
		strncat(ep->out_buf, tmp, newrd);
		rd += newrd;
	    }
	}
	ep->out_buf[ep->out_buf_size] = '\0';
	if ( echo ) printf("\n<<<OCTAVE>>>:\n%s\n", ep->out_buf); /* echo the output buffer */
    }

    return rtn;	
}

/* the array is sent as a sequence:
   struct header
   double data (if complex, first all real parts, then all imag parts)
*/     
extern PyArrayObject *engGetArray(Engine *ep, const char *name)
{
    int childpid, echo_stat;
    FILE *fp;
    char cmd[BUF_SIZE];
    PyArrayObject *ap;
    
    if ( (childpid = fork()) == -1 ) {
	perror("Can not fork");
	return 0;
    }
    else if ( childpid == 0 ) {
	sprintf(cmd, "save -mat-binary \"%s\" \"%s\"", ep->fifo, name);
        echo_stat = echo;
        echo = 0;
	if ( engEvalString(ep, cmd) != 0 ) {
	    perror("engEvalString");
	    return 0;
	}
        echo = echo_stat;
	if ( raise(SIGSTOP) != 0 ) {
	    perror("can not stop myself");
	    return 0;
	}
    }
    else {
	if ( (fp = fopen(ep->fifo, "rb")) == NULL ) {
	    perror("Can not open fifo");
	    return 0;
	}
	if ( (ap = readMatFile(fp, name)) == 0 ) {
	    perror("read");
	    return 0;
	}
	if ( waitpid(childpid, NULL, WUNTRACED) == -1 ) {
	    perror("wait");
	    return 0;
	}
	if (fclose(fp) != 0) {
	    perror("Error closing fifo");
	    return 0;
	}
	return ap;
    }
}

extern int engPutArray(Engine *ep, const char *name, PyArrayObject *ap) 
{
    int childpid, echo_stat;
    FILE *fp;
    char cmd[BUF_SIZE];

    if ( (childpid = fork()) == -1 ) {
	perror("Can not fork");
	return -1;
    }
    else if ( childpid == 0 ) {
	sprintf(cmd, "load -mat-binary -force \"%s\";", ep->fifo);
        echo_stat = echo;
        echo = 0;
	if ( engEvalString(ep, cmd) != 0 ) {
	    perror("engEvalString");
	    return -1;
	}
        echo = echo_stat;
	if ( raise(SIGSTOP) != 0 ) {
	    perror("can not stop myself");
	    return -1;
	}
    }
    else {
	fp = fopen(ep->fifo, "wb");
	if ( writeMatFile(fp, ap, name) == -1 ) {
	    perror("write");
	    return -1;
	}
	if (fclose(fp) != 0) {
	    perror("Error closing fifo");
	    return -1;
	}
	if ( waitpid(childpid, NULL, WUNTRACED) == -1 ) {
	    perror("wait");
	    return -1;
	}
	return 0;
    }
}

/* user provide the buffer space */
extern int engOutputBuffer(Engine *ep, char *buffer, int buflen) 
{
    if (ep->out_buf != NULL) free(ep->out_buf);
    ep->out_buf = buffer;
    ep->out_buf_size = buflen;
    if (ep->out_buf != NULL) {
        ep->out_buf[0] = '\0';
        echo = 1;
    } 
    return 0; 
}
